#---------------------------------------------------------------------------
# This code is written by Norbert Huber
# Institute of Materials Research, Materials Mechanics
# Helmholtz-Zentrum Geesthacht (HZG)
# Date: 03.08.2018
#
# LICENSE
# 
# This file is part of 3DOpenPoreMaterialsAnalyzer.
#
# 3DOpenPoreMaterialsAnalyzer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# Norbert Huber, Helmholtz-Zentrum Geesthacht, either version 3 of the License, or
# (at your option) any later version.
# 
# 3DOpenPoreMaterialsAnalyzer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with 3DOpenPoreMaterialsAnalyzer.
# If not, see <http://www.gnu.org/licenses/>.
#---------------------------------------------------------------------------

from math import exp
from FFNN_Lib import *

#---------------------------------------------------------------------------
# Loading ANNs from net definitions
net_z   = FFNN( "ANN3.net"   )
net_zgi = FFNN( "ANN3gi.net" )
net_E   = FFNN( "ANN3E.net"  )
net_sy  = FFNN( "ANN3sy.net" )
#---------------------------------------------------------------------------
# Coordination of bcc reference
z_bcc    = 8.0

# Factors for size scaling of genus corresponding to 
# RVE sizes N=12 (bcc, cub) and N=6 (dia, GA)
# These factors are only needed for analyzing RVE data. 
# For experimental data with large specimen (exp) gVinf_gV=1
gVinf_gV = {"bcc": 1.189, "cub": 1.142, "dia": 1.314, "GA": 1.111, "exp": 1.0 }

# In case the inner genus can not be provided, use:
gi_not_measured = -9999
#---------------------------------------------------------------------------

#---------------------------------------------------------------------------
# Master curve for macroscopic Young's modulus and yield strength
# Equation (10)
# Input zeta_tot is the total cut fraction, Eq. (11)
#---------------------------------------------------------------------------
def MC_E_sy( zeta_tot ):
    zeta_c = 0.822
    a0     = 1.55
    a1     = 2.*(a0*zeta_c-1.)
    a2     = 4.*a0/a1
    E_E0   = 0.0
    if zeta_tot > zeta_c:
        E_E0 = 0.0
    else:
        E_E0 = 1.0 - a0*zeta_tot + a1/(1.0 + exp( -a2*(zeta_tot - zeta_c)))
    return E_E0

#---------------------------------------------------------------------------
# Master curve for macroscopic Poisson's ratio
# Equation (13)
# Input zeta_tot is the total cut fraction, Eq. (11)
#---------------------------------------------------------------------------
def MC_nu(zeta_tot):
    zeta_c = 0.822
    n  = 1.75
    if zeta_tot < 0:
        print "Warning: Cut fraction < 0 "
        zeta_tot = 0.0
    if zeta_tot > zeta_c:
        nu = 0.0
    else:
        nu = ( 1.0 - (zeta_tot/zeta_c)**n )**(1./n)
    return nu

#---------------------------------------------------------------------------
# Master curve for scaled genus density
# Equation (16)
# Input z is the average coordination number, Eq. (17)
#---------------------------------------------------------------------------
def gscaled( z ):
    return z/2.-1.

#---------------------------------------------------------------------------
# Subroutine for analyzing a given structure 
# Includes Master curves and ANN analysis
# Description of Inputs: 
# N_z: List with 6 elements, which are
#          the number of junctions with coordination z, z=3,...,8
# gi:  inner genus, as defined in Sect. 4.1
#          after removal of ligaments that touch the RVE boundaries
#          computed from g=1-V+E, Eq. (S.12), with
#              V=number of vertices (junctions or nodal masses)
#              E=number of edges    (branches  or ligaments)
#      In case, gi cannot be determined set gi to gi_not_measured
#      	   and ANN3 will be used instead of ANN3gi for determination of 
#          the average coordination number
# gVinf_gV: size scaling factor for genus correction for limited RVE size
#      Values are given in Supplementary Table 3
#---------------------------------------------------------------------------
def analyzeStructure( N_z, gi, gVinf_gV ):
        
    # Analyzing data from accessible topology of a structure
    # Calculating Total of all detectable junctions with z>=3
    # Generating input data array for ANN3
    Inputz    = [] # Input for ANN3
    Inputzgi  = [] # Input for ANN3gi
    InputEs   = [] # Input for ANN3E and ANN3sy
    Imin      = 3
    NJ        = 0
    
    # Histogram of junctions for z=Imin to 8 corrsponds to list elements i=0,...,5
    for i in range(Imin,8+1):
        NJ  += N_z[i-Imin]
    # Estimated genus density calculated from detectable junctions 
    gEst3 = gi/float(NJ)*gVinf_gV 
    
    # Assigning data to input vectors to be fed into ANNs
    for i in range(Imin,8+1):
        Inputz.append( N_z[i-Imin]/float(NJ) )
        Inputzgi.append( N_z[i-Imin]/float(NJ) )
        InputEs.append( N_z[i-Imin]/float(NJ) )

    # Information about inner genus is provided
    # ANN3gi can be used for determinaiton of average coordination number
    print
    if gi != gi_not_measured: 
        # Adding estimate for scaled genus density to input definition
        # generates input for ANN3gi
        Inputzgi.append( gEst3 )        
        y_zgi          = net_zgi.calcNeuralNetwork( Inputzgi )
        z_Avg_ANN      = y_zgi[0]
        print "Results computed using ANN3gi:"

    # Otherwise ANN3 must be used for determinaiton of average coordination number
    else: 
        # Solving ANN3
        y_z            = net_z.calcNeuralNetwork( Inputz )
        z_Avg_ANN      = y_z[0]
        print "Results computed using ANN3:"

    zeta_tot_ANN       = 1.-z_Avg_ANN/z_bcc 
    genus_density_ANN  = gscaled( z_Avg_ANN )
        
    E_E0_ANN_Master    = MC_E_sy( zeta_tot_ANN )
    nu_nu0_ANN_Master  = MC_nu( zeta_tot_ANN )


    # Solving ANN3E for correction of macroscopic Young's modulus
    y_E                = net_E.calcNeuralNetwork( InputEs )
    E_E0ANN            = y_E[0] + E_E0_ANN_Master

    # Solving ANN3sy for correction of macroscopic yield strength
    y_sy               = net_sy.calcNeuralNetwork( InputEs )
    sy_sy0ANN          = y_sy[0] + E_E0_ANN_Master

    print "Topological parameters:"
    print "z_avg                      =", z_Avg_ANN
    print "zeta_tot                   =", zeta_tot_ANN
    print "g_V*                       =", genus_density_ANN
    
    print
    print "Mechanical properties:"
    print "E/E0,bcc   (master curve)  =", E_E0_ANN_Master
    print "E/E0,bcc   (ANN corrected) =", E_E0ANN
    print "sy/sy0,bcc (master curve)  =", E_E0_ANN_Master
    print "sy/sy0,bcc (ANN corrected) =", sy_sy0ANN
    print "nu/nu0,bcc (master curve)  =", nu_nu0_ANN_Master
    
#---------------------------------------------------------------------------
# Main code solving workflow depicted in Figure 8
#---------------------------------------------------------------------------
# Selected example problems from RVE data.
# Data for statistics of junctions and inner genus are taken
# from Supplementary Data files
#----------------------------------------------------------
# Example #1
# bcc, zeta_ini=0, zeta=0 (testing the reference structure)
# List of Junctions with coordination numbers of z
# z =    3       4       5       6       7       8
Structure = "bcc"
N_z = [  0,      0,      0,      0,      0,   3059]
gi  = 7722
print
print "--------------------------------------------"
print "Example #1"
print "Analyzing bcc structure, zeta_tot = 0.0"
print "--------------------------------------------"
analyzeStructure( N_z, gi_not_measured, gVinf_gV[Structure] )
analyzeStructure( N_z, gi, gVinf_gV[Structure] )

#----------------------------------------------------------
# Example #2
# bcc, zeta_ini=0.25, zeta=0 (cut fraction for cubic structure)
# List of Junctions with coordination numbers of z
# z =    3       4       5       6       7       8
Structure = "bcc"
N_z = [ 61,    259,    646,    972,    816,    292]
gi  = 5025
print
print "--------------------------------------------"
print "Example #2"
print "Analyzing bcc structure, zeta_tot = 0.25"
print "--------------------------------------------"
analyzeStructure( N_z, gi_not_measured, gVinf_gV[Structure] )
analyzeStructure( N_z, gi, gVinf_gV[Structure] )

#----------------------------------------------------------
# Example #3
# bcc, zeta_ini=0.5, zeta=0 (cut fraction for diamond structure)
# List of Junctions with coordination numbers of z
# z =    3       4       5       6       7       8
Structure = "bcc"
N_z = [691,    830,    636,    347,     85,     17 ]
gi  = 2313
print
print "--------------------------------------------"
print "Example #3"
print "Analyzing bcc structure, zeta_tot = 0.50"
print "--------------------------------------------"
analyzeStructure( N_z, gi_not_measured, gVinf_gV[Structure] )
analyzeStructure( N_z, gi, gVinf_gV[Structure] )

#----------------------------------------------------------
# Example #4
# cubic, zeta_ini=0.33, zeta=0 (cut fraction for diamond structure)
# List of Junctions with coordination numbers of z
# z =    3       4       5       6       7       8
Structure = "cub"
N_z = [406,    509,    416,    161,      0,      0 ]
gi  = 1194
print
print "--------------------------------------------"
print "Example #4"
print "Analyzing cubic structure, zeta_tot = 0.50"
print "--------------------------------------------"
analyzeStructure( N_z, gi_not_measured, gVinf_gV[Structure] )
analyzeStructure( N_z, gi, gVinf_gV[Structure] )

#----------------------------------------------------------
# Example #5
# diamond, zeta_ini = 0, zeta=0 (diamond structure)
# List of Junctions with coordination numbers of z
# z =    3       4       5       6       7       8
Structure = "dia"
N_z = [  0,   1530,      0,      0,      0,      0]
gi  = 1135
print
print "--------------------------------------------"
print "Example #5"
print "Analyzing diamond structure, zeta_tot = 0.50"
print "--------------------------------------------"
analyzeStructure( N_z, gi_not_measured, gVinf_gV[Structure] )
analyzeStructure( N_z, gi, gVinf_gV[Structure] )
#----------------------------------------------------------

#----------------------------------------------------------
# Example #6
# cubic, zeta_ini=0.50, zeta=0 (cut fraction for GA structure)
# List of Junctions with coordination numbers of z
# z =    3       4       5       6       7       8
Structure = "cub"
N_z = [499,    373,    144,     29,      0,      0 ]
gi  = 427
print
print "--------------------------------------------"
print "Example #6"
print "Analyzing cubic structure, zeta_tot = 0.625 "
print "--------------------------------------------"
analyzeStructure( N_z, gi_not_measured, gVinf_gV[Structure] )
analyzeStructure( N_z, gi, gVinf_gV[Structure] )

#----------------------------------------------------------
# Example #7
# diamond, zeta_ini = 0.1, zeta=0.2 (<= GA structure)
# List of Junctions with coordination numbers of z
# z =    3       4       5       6       7       8
Structure = "dia"
N_z = [613,    395,      0,      0,      0,      0]
gi  = 401
print
print "--------------------------------------------"
print "Example #7"
print "Analyzing diamond structure, zeta_tot = 0.64"
print "--------------------------------------------"
analyzeStructure( N_z, gi_not_measured, gVinf_gV[Structure] )
analyzeStructure( N_z, gi, gVinf_gV[Structure] )
#----------------------------------------------------------

