# -*- coding: cp1252 -*-
#---------------------------------------------------------------------------
# This code is written by Norbert Huber
# Institute of Materials Research, Materials Mechanics
# Helmholtz-Zentrum Geesthacht (HZG)
# Date: 03.08.2018
#
# LICENSE
# 
# This file is part of 3DOpenPoreMaterialsAnalyzer.
#
# 3DOpenPoreMaterialsAnalyzer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# Norbert Huber, Helmholtz-Zentrum Geesthacht, either version 3 of the License, or
# (at your option) any later version.
# 
# 3DOpenPoreMaterialsAnalyzer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with 3DOpenPoreMaterialsAnalyzer.
# If not, see <http://www.gnu.org/licenses/>.
#---------------------------------------------------------------------------
import math, random

#Global constants
epsilon = 1.e-3;
nMinIn  = -0.25
nMaxIn  = 0.25
nMinOut = 0.25
nMaxOut = 0.75

#Global functions
def sigmoidal( v ):
    return 1./(1.+math.exp(-v))

#Neuron class definition
class Neuron(object):
    def __init__(self, L, N, I_in):
        self.L = L
        self.N = N
        self.I_in = I_in
        self.w = []
        self.v = 0.0
        self.y = -1.0

    def setWeight( self, w ):
        self.w.append(w)

    def result( self, x ):
        self.v = 0.0
        for i in range(0,len(x)):
            self.v += x[i]*self.w[i]
        self.y = sigmoidal(self.v)
        return self.y

    
#Layer class definition
class Layer(object):
    def __init__(self, L, K, I):
        self.L = L
        self.N_list = []
        self.K = K
        self.I = I
        self.y = []

        for k in range(self.K+1):
            self.N_list.append( Neuron( self.L, k, self.I ) )
            self.y.append( -1.0 )

    def setLayer( self, w ):
        for k in range(0,self.K):
            self.N_list[k].set( w[(self.I+1)*k] )

    def calcLayer( self, x ):
        for k in range(0,self.K):
            self.y[k] = self.N_list[k].result( x )

#Network class definition
class FFNN(object):

    def __init__(self, filename):
        self.IErrIn = []
        self.IErrOut = []
        self.inputs = []
        self.outputs = []
        self.y = []
        self.d = []
        self.name = filename.split( ".", 1 )[0]

        print "Loading neural network", self.name
        in_file = open( filename, "r" )        
        line = ""
        
        #read number of layers and assign value to ANN
        while line.find("L = ") < 0:
            line = in_file.next()
        L = int(line.split("=")[1])
        self.N_Layers = int(L)
        print "Number of layers:", self.N_Layers
        
        #read number of neurons per layer and assign values to ANN
        while line.find("NL = ") < 0:
            line = in_file.next()
        NLstring = line[line.find("[")+1:line.find("]")].split(",")
        
        self.N_Neurons = []
        for i in range(0,self.N_Layers):
              self.N_Neurons.append(int(NLstring[i]))
        print "Number of neurons:", self.N_Neurons
        N_in  = self.N_Neurons[0]
        N_out = self.N_Neurons[-1]
        
        #calculate number of synaptic weights
        self.numberWeights = self.countWeights()
        print "Number of synaptic weights:", self.numberWeights        

        #read synaptic weights and assign values to ANN
        while line.find("W = ") < 0:
            line = in_file.next()
        Wstring = line[line.find("[")+1:line.find("]")].split(",")
        NWread  = len(Wstring)
        if NWread != self.numberWeights:
            print "Error in reading synaptic weights. Number is not consistent!"
            exit()
        
        self.W = []
        for i in range(0,self.numberWeights):
              self.W.append(float(Wstring[i]))
        print "End reading %d synaptic weights." % self.numberWeights

        #read min-max values for input neurons
        while line.find("min_in = ") < 0:
            line = in_file.next()
        Mstring = line[line.find("[")+1:line.find("]")].split(",")

        if len(Mstring)!= N_in:
            print "Error in reading min values for inputs. Number is not consistent!"
            exit()


        self.minIn = []
        for i in range(0,self.N_Neurons[0]):
            self.minIn.append(float(Mstring[i]))

        while line.find("max_in = ") < 0:
            line = in_file.next()
        Mstring = line[line.find("[")+1:line.find("]")].split(",")

        if len(Mstring)!= N_in:
            print "Error in reading max values for inputs. Number is not consistent!"
            exit()

        self.maxIn = []
        for i in range(0,self.N_Neurons[0]):
            self.maxIn.append(float(Mstring[i]))
            self.inputs.append( 0.0 )
            self.IErrIn.append( 0.0 )            

        while line.find("min_out = ") < 0:
            line = in_file.next()
        Mstring = line[line.find("[")+1:line.find("]")].split(",")

        if len(Mstring)!= N_out:
            print "Error in reading min values for outputs. Number is not consistent!"
            exit()

        self.minOut = []
        for i in range(0,self.N_Neurons[-1]):
            self.minOut.append(float(Mstring[i]))
            self.outputs.append( 0.0 )
            self.y.append( 0.0 )            
            self.d.append( 0.0 )            
            self.IErrOut.append( 0.0 )            

        while line.find("max_out = ") < 0:
            line = in_file.next()
        Mstring = line[line.find("[")+1:line.find("]")].split(",")

        if len(Mstring)!= N_out:
            print "Error in reading max values for outputs. Number is not consistent!"
            exit()

        self.maxOut = []
        for i in range(0,self.N_Neurons[-1]):
            self.maxOut.append(float(Mstring[i]))

        #close input and set up neural network
        in_file.close()
        self.setUpFFNN()

    def countWeights(self):
        numberWeights = 0
        for n in range(1,self.N_Layers):
            for k in range(0,self.N_Neurons[n]):
                for i in range(0,self.N_Neurons[n-1]+1):
                    numberWeights += 1
        return numberWeights
        

    def setUpFFNN(self):
        print "Setting up neural network."
        print "Defining %d network layers." % self.N_Layers
        self.L_pointer = [Layer(0, self.N_Neurons[0], self.N_Neurons[0])]
        for l in range(1, len(self.N_Neurons)):
            self.L_pointer.append(Layer(l, self.N_Neurons[l], self.N_Neurons[l-1]))

        print "Assigning %d synaptic weights" % self.numberWeights, "to neural network."
        iw = 0
        for n in range(1,self.N_Layers):
            for k in range(0,self.N_Neurons[n]):
                for i in range(0,self.N_Neurons[n-1]+1):
                    #print "Layer", n-1, "Neuron", i, "Layer", n, "Neuron", k, "Weight", iw, self.W[iw]
                    self.L_pointer[n].N_list[k].setWeight( self.W[iw] )
                    iw += 1
        print "Neural network", self.name, "is ready for calculation."

    def scaleInputs(self, x):
        for i in range(0,self.N_Neurons[0]):
            if x[i] < self.minIn[i]:
                self.IErrIn[i] = -1
            elif x[i] > self.maxIn[i]:
                self.IErrIn[i] = 1
            else:
                self.IErrIn[i] = 0
            self.inputs[i] = (x[i]-self.minIn[i])/(self.maxIn[i]-self.minIn[i])*(nMaxIn-nMinIn)+nMinIn

    def feedForward(self):
        l = 0
        for i in range(0,self.N_Neurons[l]):
            self.L_pointer[l].y[i] = self.inputs[i]
            self.L_pointer[l].N_list[i].y = self.inputs[i]
        for l in range(1,self.N_Layers):
            self.L_pointer[l].calcLayer(self.L_pointer[l-1].y)

        for i in range(0,self.N_Neurons[-1]):
            self.outputs[i] = self.L_pointer[-1].y[i]
            
    def unscaleOutputs(self):
        for i in range(0,self.N_Neurons[-1]):
            self.y[i] = (self.outputs[i]-nMinOut)/(nMaxOut-nMinOut)*(self.maxOut[i]-self.minOut[i]) + self.minOut[i]

            if self.y[i] < self.minOut[i]:
                self.IErrOut[i] = -1
            elif self.y[i] > self.maxOut[i]:
                self.IErrOut[i] = 1
            else:
                self.IErrOut[i] = 0

        return self.y;

    def calcNeuralNetwork(self, x):
        self.scaleInputs(x)
        self.feedForward()
        y = self.unscaleOutputs()
        return y


